package be.iminds.ilabt.jfed.experimenter_gui.ui;

import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.util.SentryUtil;
import be.iminds.ilabt.jfed.module.JFedWebApiClientModule;
import be.iminds.ilabt.jfed.bugreport.model.BugReport;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportComposer;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportToServerTask;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClientConfig;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.common.ExceptionUtil;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import com.google.common.base.Throwables;
import com.google.inject.name.Named;
import io.sentry.Sentry;
import javafx.application.Platform;
import javafx.scene.control.ButtonType;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.layout.VBox;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.stage.Stage;
import org.controlsfx.dialog.ExceptionDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import java.util.Objects;
import java.util.Optional;

import static be.iminds.ilabt.jfed.highlevel.util.SentryUtil.createSentrySlices;
import static be.iminds.ilabt.jfed.util.sentry.SentryUtil.createSentryPreferences;

/**
 * User: twalcari
 * Date: 6/25/2015
 * Time: 13:48
 */
public class UncaughtGuiExceptionToBugReport implements Thread.UncaughtExceptionHandler {
    private static final Logger LOG = LoggerFactory.getLogger(UncaughtGuiExceptionToBugReport.class);

    private final Provider<BugReportComposer> bugReportComposerProvider;
    private final Provider<FedmonWebApiClientConfig> fedmonWebApiClientConfigProvider;
    private final Provider<TaskService> taskServiceProvider;
    private final SfaModel sfaModel;
    private final JFedPreferences jFedPreferences;
    private Stage stage;

    @Inject
    public UncaughtGuiExceptionToBugReport(
            Provider<BugReportComposer> bugReportComposerProvider,
            @Named(JFedWebApiClientModule.NAMED__CLIENT_AUTH_WEBAPI_CONFIG) Provider<FedmonWebApiClientConfig> fedmonWebApiClientConfigProvider,
            Provider<TaskService> taskServiceProvider,
            SfaModel sfaModel,
            JFedPreferences jFedPreferences) {
        this.bugReportComposerProvider = bugReportComposerProvider;
        this.fedmonWebApiClientConfigProvider = fedmonWebApiClientConfigProvider;
        this.taskServiceProvider = taskServiceProvider;
        this.sfaModel = sfaModel;
        this.jFedPreferences = jFedPreferences;
    }

    public void setStage(Stage stage) {
        this.stage = stage;
    }

    @Override
    public void uncaughtException(Thread t, Throwable uncaughtThrowable) {
        LOG.error("UNCAUGHT ERROR IN THREAD {}", t.getName(), uncaughtThrowable);

        try {
            for (StackTraceElement ste : uncaughtThrowable.getStackTrace()) {
                if (ste.getClassName().contains(UncaughtGuiExceptionToBugReport.class.getSimpleName())) {
                    LOG.error("DETECTED CYCLE IN UNCAUGHT ERROR. WILL ABORT!", uncaughtThrowable);
                    return;
                }
            }
        } catch (Throwable cycleDetectEx) {
            LOG.error("ERROR IN CYCLE DETECTION. WILL ABORT!", cycleDetectEx);
            return;
        }

        Platform.runLater(() -> {
            try {
                if (ExceptionUtil.isOutOfMemory(uncaughtThrowable)) {
                    if (stage != null) {
                        try {
                            Optional<ButtonType> reply =
                                    JFDialogs.create()
                                            .owner(stage)
                                            .message("jFed has run out of memory. This will cause bugs and instabilities.\n" +
                                                    "It is advised to close jFed now and restart it.\n\n" +
                                                    "You can choose to continue. You might still be able to save your work.\n\n" +
                                                    "Continue running jFed?")
                                            .masthead("Out of Memory!")
                                            .buttonTypes(ButtonType.YES, ButtonType.NO, ButtonType.CANCEL)
                                            .showWarning();

                            if (reply.isPresent() && Objects.equals(reply.get(), ButtonType.YES)) {
                                LOG.warn("User chose to continue after out of memory.");
                            } else {
                                LOG.warn("User chose to exit after out of memory.");
                                System.exit(-1);
                            }
                        } catch (Throwable throwable) {
                            System.err.println("jFed encountered an exception trying to handle OutOfMemoryError dialog. Will exit.");
                            LOG.error("jFed encountered an exception trying to handle OutOfMemoryError dialog. Will exit.", throwable);
                            System.exit(-1);
                        }
                    } else {
                        System.err.println("jFed ran out of memory and will exit.");
                        LOG.warn("jFed ran out of memory and will exit.", uncaughtThrowable);
                        System.exit(-1);
                    }
                } else {
                    if (ExceptionUtil.isOutOfDiskSpace(uncaughtThrowable)) {
                        try {
                            JFDialogs.create()
                                    .owner(stage)
                                    .message("You ran out of disk space. This caused a jFed action to fail.\n" +
                                            "It is advised to free up disk space before continuing to use jFed.\n")
                                    .masthead("Out of Disk Space!")
                                    .buttonTypes(ButtonType.OK)
                                    .showWarning();
                        } catch (Throwable throwable) {
                            System.err.println("jFed encountered an exception trying to handle out of disk space dialog. Will ignore.");
                            LOG.error("jFed encountered an exception trying to handle out of disk space dialog. Will ignore.", throwable);
                        }
                    } else {
                        try {

                            ExceptionDialog exceptionDialog = new ExceptionDialog(uncaughtThrowable);
                            if (stage != null)
                                exceptionDialog.initOwner(stage);
                            exceptionDialog.setHeaderText("Unexpected error");

                            VBox contentPane = new VBox(10);
                            contentPane.setPrefWidth(600);
                            Label sorryLabel = new Label("We're sorry");
                            sorryLabel.setFont(Font.font(Font.getDefault().getFamily(), FontWeight.BOLD, Font.getDefault().getSize()));

                            Label infoLabel = new Label("jFed detected an unexpected error, which can likely cause bugs and instabilities. If you experience these, we recommend restarting jFed.");
                            infoLabel.setWrapText(true);
                            CheckBox sendBugreportCheckbox = new CheckBox("Tell jFed developers about this error, so they can fix it");
                            sendBugreportCheckbox.setWrapText(true);
                            sendBugreportCheckbox.setSelected(true);

                            TextArea descriptionTextArea = new TextArea();
                            descriptionTextArea.setPromptText("Optional: add extra info");
                            descriptionTextArea.setStyle("-fx-margin: 0 0 0 30px");
                            descriptionTextArea.setPrefRowCount(3);

                            contentPane.getChildren().addAll(sorryLabel, infoLabel, sendBugreportCheckbox, descriptionTextArea);
                            exceptionDialog.getDialogPane().setContent(contentPane);

                            Platform.runLater(() -> {
                                if (exceptionDialog.showAndWait().isPresent() && sendBugreportCheckbox.isSelected()) {
//                                    BugReportComposer bugReportComposer = bugReportComposerProvider.get();
//                                    bugReportComposer.setFirstRunConnectivityTests(false);
//
//                                    String subject = IOUtils.exceptionToShortDescription(uncaughtThrowable);
//
//                                    bugReportComposer.setPostOnPublicList(false);
//                                    bugReportComposer.setReportType(BugReport.ReportType.UNCAUGHT_EXCEPTION);
//                                    bugReportComposer.setReportTarget(BugReport.ReportTarget.JFED);
//                                    bugReportComposer.setSubject(subject);
////                            bugReportComposer.setSubject("Uncaught Exception: " + uncaughtThrowable.getMessage());
//
//                                    StringBuilder description = new StringBuilder();
//                                    description.append(descriptionTextArea.getText().trim());
//
//                                    if (description.length() > 0) {
//                                        description.append("\n---\n\n");
//                                    }
//
//                                    description
//                                            .append("Exception:").append(Throwables.getStackTraceAsString(uncaughtThrowable))
//                                            .append("\nThread: ").append(t.getName());
//
//                                    bugReportComposer.setDescription(description.toString());
//
//                                    taskServiceProvider.get().submitTask(new BugReportToServerTask(bugReportComposer, fedmonWebApiClientConfigProvider.get()));

                                    //hint: CTRL-F4 to test uncaught exception
                                    if (descriptionTextArea.getText().trim().length() >= 2) {
                                        //Sentry java API does not yet have user feedback support.
                                        //  see also:
                                        //     https://docs.sentry.io/enriching-error-data/user-feedback/
                                        //     https://docs.sentry.io/api/projects/post-project-user-reports/
                                        //Instead of using that (requires manually talking to the API, and including the correct event ID),
                                        //we'll just add "extra" with the feedback
                                        Sentry.getContext().addExtra("user_feedback_comment", descriptionTextArea.getText());
                                        Sentry.getContext().addExtra("slices", createSentrySlices(sfaModel));
                                        Sentry.getContext().addExtra("preferences", createSentryPreferences(jFedPreferences));
                                        Sentry.capture(uncaughtThrowable);
                                        Sentry.getContext().getExtra().remove("user_feedback_comment");
                                    } else {
                                        Sentry.capture(uncaughtThrowable);
                                    }
                                }
                            });


                        } catch (Throwable throwable) {
                            System.err.println("jFed encountered an exception trying to handle Error dialog. (inner)");
                            LOG.error("jFed encountered an exception trying to handle Error dialog. (inner)", throwable);
                        }
                    }
                }
            } catch (Throwable throwable) {
                Sentry.capture(throwable);
                System.err.println("jFed encountered an exception trying to handle Error dialog. (outer)");
                LOG.error("jFed encountered an exception trying to handle Error dialog. (outer)", throwable);
            }
        } );
    }


}
